/* $Id $
	------------------------------------------------------------------
	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
	------------------------------------------------------------------

	@author Nicklas
*/
'use strict';

var PlaceOnPlate = function()
{
	var pop = {};
	
	var graphics;
	var pen;
	var selectedPen;

	var plate;
	var selectedWell;
	var itemList;
	var selectedItem;

	/**
		Initialize the page.
	*/
	pop.initPage = function()
	{
		// Save and close buttons
		Buttons.addClickHandler('close', App.closeWindow);
		Buttons.addClickHandler('btnSave', pop.placeOnPlate);
		
		// Protocol
		Buttons.addClickHandler('protocol_id.select', Wizards.selectProtocol);
		Events.addEventHandler('protocol_id', 'base-selected', Wizards.setProtocolCallback);

		// Kit
		Buttons.addClickHandler('kit_id.select', Wizards.selectKit);
		Events.addEventHandler('kit_id', 'base-selected', Wizards.setKitCallback);

		// Hardware
		Buttons.addClickHandler('hardware_id.select', Wizards.selectHardware);
		Events.addEventHandler('hardware_id', 'base-selected', Wizards.setHardwareCallback);

		// Toolbar
		Buttons.addClickHandler('btnSelectPlate', pop.selectBioPlate);
		Events.addEventHandler('btnSelectPlate', 'base-selected', pop.setBioPlateCallback);
		Buttons.addClickHandler('btnPlaceByRow', pop.placeByRow);
		Buttons.addClickHandler('btnPlaceByColumn', pop.placeByColumn);
		Buttons.addClickHandler('btnClearMapping', pop.clearMapping);
		
		// Initialize graphics
		graphics = new jsGraphics(Doc.element('canvas'));
		pen = new jsPen(new jsColor('#2288AA'), 1);
		selectedPen = new jsPen(new jsColor('#2288AA'), 2);
		
		// Create the list of items
		itemList = new ItemList('itemlist');
		var items = Doc.element('itemlist').getElementsByClassName('item');
		for (var i = 0; i < items.length; i++)
		{
			var item = items[i];
			var itemId = Data.int(item, 'item-id');
			var name = Data.get(item, 'name');
			if (itemId) 
			{
				itemList.addItem(itemId, name);
				Events.addEventHandler(item, 'click', pop.itemOnClick);
				Events.addEventHandler(item, 'mouseover', pop.itemOnMouseOver);
				Events.addEventHandler(item, 'mouseout', pop.itemOnMouseOut);
			}
		}
		// Handle layout changes
		Events.addEventHandler('itemlist', 'scroll', pop.onLayoutChange);
		Events.addEventHandler(window, 'resize', pop.onLayoutChange);
	}
	
	pop.selectBioPlate = function()
	{
		if (plate && !confirm('This will remove all placed biomaterial. Continue?')) return;
		
		var url = '&resetTemporary=1';
		var bioMaterialType = Data.int('page-data', 'biomaterial-type');
		url += '&tmpfilter:INT:bioPlateType.bioMaterialType='+encodeURIComponent('|'+bioMaterialType);
		var commonSubtype = Data.int('page-data', 'common-subtype');
		if (commonSubtype)
		{
			url += '&tmpfilter:INT:bioPlateType.itemSubtype='+encodeURIComponent('|'+commonSubtype);
		}
		Dialogs.selectItem('BIOPLATE', 'btnSelectPlate', 0, url);
	}

	pop.setBioPlateCallback = function(event)
	{
		var request = Ajax.getXmlHttpRequest();
		var url = '../bioplates/ajax.jsp?ID='+App.getSessionId()+'&cmd=GetFullPlateInfo&item_id=' + event.detail.id;
		request.open("GET", url, false); 
		// NOTE! 'false' causes code to wait for the response. aka. 'Synchronous AJAX' or SJAX.
		request.send(null);

		var plateInfo = JSON.parse(request.responseText);
		if (plateInfo.status != 'ok')
		{
			Forms.showNotification('plate.name', plateInfo.message);
			return false;
		}
		
		// Get plate and well information from the AJAX response
		var rows = plateInfo.rows;
		var columns = plateInfo.columns;
		
		var bigPlate = rows > 12 || columns > 18;
		var plateClass = bigPlate ? 'plate bigplate' : 'plate';

		// Remove existing mapping
		itemList.unmapAll(graphics);

		// Create plate and wells
		plate = new Plate('dest', plateInfo.id, plateInfo.name, rows, columns);
		for (var i = 0; i < plateInfo.wells.length; i++)
		{
			var wellInfo = plateInfo.wells[i];
			var bmInfo = wellInfo.bioMaterial;
			var row = parseInt(wellInfo.row);
			var col = parseInt(wellInfo.column);
			var well = plate.getWell(row, col);
			well.locked = !wellInfo.canAdd || bmInfo;
			well.name = bmInfo ? bmInfo.name : '';
		}
		
		Doc.element('plate.name').innerHTML = Strings.encodeTags(plateInfo.name);
		
		// Create html table representing the bioplate
		var html = '<table class="'+plateClass+'">';
		html += '<tr><td class="bg-filled-100"></td>';
		for (var c = 0; c < plate.columns; c++)
		{
			html += '<td class="columnheader bg-filled-100">' + (c+1) + '</td>';
		}
		html += '</tr>';
		for (var r = 0; r < plate.rows; r++)
		{
			html += '<tr><td class="rowheader bg-filled-100">' + Plates.toAlphaCoordinate[r] + '</td>';
			for (var c = 0; c < plate.columns; c++)
			{
				var well = plate.getWell(r, c);
				var cls = 'well';
				var title = '';
				if (well.locked) 
				{
					cls+= ' unmappable';
					if (well.name) 
					{
						title = 'This well is used by: ' + Strings.encodeTags(well.name);
					}
					else
					{
						title = 'This well has already been used';
					}
				}
				else
				{
					cls += ' empty editable';
				}
				html += '<td id="dest.'+r+'.'+c+'" class="' + cls + '" data-row="'+r+'" data-column="'+c+'" title="'+title+'"></td>';
			}
			html += '</tr>';
		}
		html += '</table>';
		Doc.element('plate').innerHTML = html;
		
		var wells = Doc.element('plate').getElementsByClassName('editable');
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			Events.addEventHandler(well, 'click', pop.wellOnClick);
			Events.addEventHandler(well, 'mouseout', pop.wellOnMouseOut);
			Events.addEventHandler(well, 'mouseover', pop.wellOnMouseOver);
		}
		
		Doc.show('toolbar.mappings');
	}

	
	pop.placeOnPlate = function()
	{
		if (!plate)
		{
			Forms.showNotification('btnSelectPlate', 'No plate has been selected');
			return;
		}
		var frm = document.forms['main'];
		if (Strings.trim(frm.name.value) == '')
		{
			Forms.showNotification(frm.name, 'You must enter a name for the event');
			return;
		}
		
		frm.plate_id.value = plate.id;
		frm.rows.value = plate.rows;
		frm.columns.value = plate.columns;
		
		var numMapped = 0;
		for (var row = 0; row < plate.rows; row++)
		{
			for (var column = 0; column < plate.columns; column++)
			{
				var well = plate.getWell(row, column);
				
				if (well.mappedItem)
				{
					Forms.addHidden(frm, 'well.' + row + '.' + column, well.mappedItem.id);
					numMapped++;
				}
			}
		}
		
		if (numMapped == 0)
		{
			Forms.showNotification('btnSave', 'No biomaterial has been mapped to wells');
			return;
		}
		
		frm.submit();
	}
	
	/**
		Draw a link between the item and it's mapped well.
	*/
	pop.itemOnMouseOver = function(event)
	{
		var itemId = Data.get(event.currentTarget, 'item-id');
		var item = itemList.getItem(itemId);
		item.drawLink(graphics, pen, false);
	}
	
	/**
		Hide the link between the item and it's mapped well, 
		unless the item (or it's mapped well) is selected.
	*/
	pop.itemOnMouseOut = function(event)
	{
		var itemId = Data.get(event.currentTarget, 'item-id');
		var item = itemList.getItem(itemId);

		if (!item.selected && !(item.mappedWell && item.mappedWell.selected)) 
		{
			item.hideLink(graphics);
		}
	}

	/**
		Select the item that is clicked on. The currently selected item
		is de-selected. If the clicked item is the same as the currently 
		selected item no new item is selected. If a well is already selected
		a link is made between the well and the newly selected item.
	*/
	pop.itemOnClick = function(event)
	{
		var itemId = Data.get(event.currentTarget, 'item-id');
		var item = itemList.getItem(itemId);
	
		// De-select the currently selected item
		if (selectedItem) 
		{
			selectedItem.setSelected(false);
			if (item == selectedItem)
			{
				// Re-draw link with regular pen (eg. same as onmouseover)
				selectedItem.drawLink(graphics, pen, true);
				selectedItem = null;
				return;
			}
			else
			{
				selectedItem.hideLink(graphics);
			}
		}
		
		// Select the new item and draw link to mapped well
		selectedItem = item;
		selectedItem.setSelected(true);
		selectedItem.drawLink(graphics, selectedPen, true);
	
		// Map the item to the selected well
		if (selectedWell && selectedItem)
		{
			pop.placeSelectedOnPlate();
		}
	}

	
	/**
		Draw a link between the well and it's mapped item.
	*/
	pop.wellOnMouseOver = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var column = Data.int(event.currentTarget, 'column');

		var well = plate.getWell(row, column);
		if (!well || !well.mappedItem) return;
		well.mappedItem.scrollIntoView();
		well.mappedItem.drawLink(graphics, pen, false);
	}
	
	/**
		Hide the link between the well and it's mapped item, 
		unless the well (or it's mapped item) is selected.
	*/
	pop.wellOnMouseOut = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var column = Data.int(event.currentTarget, 'column');

		var well = plate.getWell(row, column);
		if (!well || !well.mappedItem) return;
		if (!well.selected && !(well.mappedItem && well.mappedItem.selected)) 
		{
			well.hideLink(graphics);
		}
	}
	
	/**
		Select the well that is clicked on. The currently selected well is
		de-selected. If the clicked well is the same as the currently 
		selected well no new well is selected. If an item is already selected
		a link is made between the item and the newly selected well.
	*/
	pop.wellOnClick = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var column = Data.int(event.currentTarget, 'column');
		var well = plate.getWell(row, column);
	
		// De-select the currently selected well
		if (selectedWell) 
		{
			selectedWell.setSelected(false);
			if (well == selectedWell)
			{
				// Re-draw link with regular pen (eg. same as onmouseover)
				selectedWell.drawLink(graphics, pen, true);
				selectedWell = null;
				return;
			}
			else
			{
				selectedWell.hideLink(graphics);
			}
		}
		
		selectedWell = well;
		selectedWell.setSelected(true);
		selectedWell.drawLink(graphics, selectedPen, true);
	
		if (selectedWell && selectedItem)
		{
			var nextItem = Doc.element('autoSelectUnmappedItem').checked ? 
					itemList.nextUnmappedItem(selectedItem) : null;
			pop.placeSelectedOnPlate();
			if (nextItem)
			{
				nextItem.setSelected(true);
				nextItem.scrollIntoView();
				selectedItem = nextItem;
			}
		}
	}
	
	/**
		Redraw the link to the selected item/well when the item list scrolls.
	*/
	pop.onLayoutChange = function(event)
	{
		var item = selectedItem;
		if (!item && selectedWell) item = selectedWell.mappedItem;
		if (item) item.drawLink(graphics, selectedPen, true);
	}


	/**
		Map the selected item with the selected well. Hide and
		redraw links as needed.
	*/
	pop.placeSelectedOnPlate = function()
	{
		// Hide any links that are currently displayed
		selectedItem.hideLink(graphics);
		selectedWell.hideLink(graphics);
		
		// Map to the new well and draw link
		if (selectedItem.mappedWell != selectedWell)
		{
			selectedItem.mapToWell(selectedWell);
			selectedItem.scrollIntoView();
			selectedItem.drawLink(graphics, pen, true);
		}
		else
		{
			selectedItem.unmapWell();
		}
		
		// De-select everything
		selectedItem.setSelected(false);
		selectedWell.setSelected(false);
		selectedItem = null;
		selectedWell = null;
	}

	/**
		Remove all mappings that have been made so far.
	*/
	pop.clearMapping = function()
	{
		if (!plate)
		{
			Forms.showNotification('btnSelectPlate', 'No destination plate has been selected');
			return;
		}
		if (!confirm('This will remove all placed biomaterial. Continue?')) return;
		itemList.unmapAll(graphics);
		if (selectedItem)
		{
			selectedItem.setSelected(false);
			selectedItem = null;
		}
		if (selectedWell)
		{
			selectedWell.setSelected(false);
			selectedWell = null;
		}
	}

	/**
		Automatically place unplaced biomaterial on the plate filling rows first.
		Biomaterial and wells that have already been mapped are skipped.
	*/
	pop.placeByRow = function()
	{
		if (!plate)
		{
			Forms.showNotification('btnSelectPlate', 'No destination plate has been selected');
			return;
		}
		var index = 0;
		var numItems = itemList.items.length;
		for (var row = 0; row < plate.rows; row++)
		{
			for (var column = 0; column < plate.columns; column++)
			{
				var well = plate.getWell(row, column);
				if (!well.mappedItem && !well.locked)
				{
					var mapped = false;
					while (index < itemList.items.length && !mapped)
					{
						var item = itemList.items[index];
						index++;
						if (!item.mappedWell)
						{
							item.mapToWell(well);
							mapped = true;
						}
					}
				}
			}
		}
	}
	
	/**
		Automatically place unplaced biomaterial on the plate filling columns first.
		Biomaterial and wells that have already been mapped are skipped.
	*/
	pop.placeByColumn = function()
	{
		if (!plate)
		{
			Forms.showNotification('btnSelectPlate', 'No destination plate has been selected');
			return;
		}
		var index = 0;
		var numItems = itemList.items.length;
		for (var column = 0; column < plate.columns; column++)
		{
			for (var row = 0; row < plate.rows; row++)
			{
				var well = plate.getWell(row, column);
				if (!well.mappedItem && !well.locked)
				{
					var mapped = false;
					while (index < itemList.items.length && !mapped)
					{
						var item = itemList.items[index];
						index++;
						if (!item.mappedWell)
						{
							item.mapToWell(well);
							mapped = true;
						}
					}
				}
			}
		}
	}

	return pop;
}();

Doc.onLoad(PlaceOnPlate.initPage);
